/*---------------------------------------------------------------------------------------

 Halls 2015. Post-Vega
---------------------------------------------------------------------------------------*/
; Change this to point somewhere suitable on your computer.
; Note that Windows 7 doesn't let you write into the root directory!

OutputDir               equ "..\\bin"                     ; This is where we'll put the files

                        export_sym "halls.sym",0

; Some testing options

;bFixedMaze              equ true                        ; Same one every time
;bMyDefaults             equ true                        ; For testing
;bReturnInstantResult    equ true                        ; For testing
;bNoDamage               equ true                        ; Don't damage the man
;bRobsKeys               equ true                        ; Use Rob's preferred keys
bCompressForTape        equ true                        ; Compress the file

                        include "equ.inc"               ; Equates

                        emulate_spectrum                ; Tell the emulator what we are

; Our physical keycodes

        struct
pkNone                  ds 1                            ;
pkUp                    ds 1                            ;
pkDown                  ds 1                            ;
pkLeft                  ds 1                            ;
pkRight                 ds 1                            ;
pkFire                  ds 1                            ;
pkSelect                ds 1                            ;
pk1                     ds 1                            ;
nPhysicalKeys send

        struct
dKeyAddr                ds 2
dKeyName                ds 1
dKeyPressed             ds 1
dKeyReport              ds 1
dKeyCount               ds 1
dKeyRepeat              ds 1
lKeyEntry send

KEY_DEB                 equ 2                           ; Debounce count
iKEY_REPEAT_CNT         equ 30
KEY_REPEAT_CNT          equ 6

; Menu definitions

        struct
dMenuLine               ds 2
dMenuUp                 ds 2
dMenuDown               ds 2
dMenuExec               ds 2
dMenuFlags              ds 1
        send

; Macro for defining menus

mDefineMenu             macro(Line,Up,Down,Exec,Flags)
                          dw Line*32+#5800,Up,Down,Exec
                          db Flags
                          mend

; Start of Spectrum free memory

                        org $5B00                       ; Immediately after the screen

                        ; Include the fixed data and the start vector

                        include "chars.inc"             ; The various character sets

MaxHighScoreText        equ 104                         ; No more than this many characters
MaxHighScoreBytes       equ (MaxHighScoreText + 3)*16   ; Allow for some

; A macro to declare highscore table entries

hse                     macro (str,score)
                          db str,0
                          dw score
                          mend

Front_Vars_Start        equ *

HighScores              hse("Martin Horsley - The Good",8085)                   ; He really was, nobody could touch him at playing halls.
                        hse("Neil Mottershead - The Bad",8080)                  ;
                        hse("Simon Brattel - The Ug... what? Oi!",6809)         ;
                        hse("Graham Stafford - the betrayer!",6502)             ; He asked for this. Search me...
                        hse("Matt Smith - no, not that one...",6502)            ; He wrote a PC retro remake of halls
                        hse("Andy Stewart - (beautiful remake)",6502)           ; He wrote a perfect PC clone. Love it.
                        hse("Rob Crawford - Loader? What loader?",6502)         ; He's, ah, reading this right now. Hello...
                        hse("Shaun McClure - How many pixels???",6502)          ; The dead man graphic and a thing
                        hse("John Marshall - For 'The Keys of Krall'",6502)     ; To add this or not to add this?
                        hse(" ",4004)                                           ; Need some lines
                        hse(cBlurb,1802)                                        ; Should have thought of this years ago
;                        hse(" ",4004)                                           ; Need some lines

                        ; Special case entries here
                        db "You killed ",cnThings," of the things? Gah...",0
                        dw 1654

                        db cScore," point",cAnEss," - call that a score?",0
                        dw 1654

                        hse(" ",1)
                        hse("For the love of all that's lost",1)
;                        hse(" ",1)
                        hse("And now for something completely the same...",1)

HighScoreDataEnd        equ HighScores+MaxHighScoreBytes;

                        ; Clear the gap up to the blurbs

        if true // This breaks the new assembler. I wonder why?
                        while . < (HighScoreDataEnd - (sBlurbEnd - sBlurbStart))
                          db ZEUSRAND and $FF
                          wend
        else
                        ds (HighScoreDataEnd - (sBlurbEnd - sBlurbStart)) - .,$00
        endif

                        ; We want to use this space for extra messages

nBlurbs                 = 0

blb                     macro(str)
                          dbx *,str,0
                          ::nBlurbs = ::nBlurbs+1
                          mend

sBlurbStart             equ .

BlurbValidMarker        db $1F                          ; A rough check we haven't been overwritten

sBlurbs                 blb("Who did he betray?")
                        blb("Hello? Why am I bad?")
                        blb("Bad at what, exactly?")
                        blb("I'm confused...")
                        blb("Something changed...")
                        blb("Hang on...")

sBlurbEnd               equ .

                        org HighScoreDataEnd

nSomeone                db 0                            ;
nBlurb                  db 0                            ;

sHighC                  defb $07
HSChar                  defb "1",$92,$03,$00

pMainMenu               ds 2                            ;

; The keyboard debounce table

mDefineKeyCount         macro(KeyAddr,Name)
                          dw KeyAddr
                          db Name
                          db 0,0,0,0
                          mend

PhysicalKeyTable        mDefineKeyCount(KeyLeft,pkLeft)
                        mDefineKeyCount(KeyRight,pkRight)
                        mDefineKeyCount(KeyUp,pkUp)
                        mDefineKeyCount(KeyDown,pkDown)

                        mDefineKeyCount(KeyLight,pkFire)
                        mDefineKeyCount(KeyJustify,pkSelect)
                        mDefineKeyCount(KeyFire,pk1)

                        dw 0

Front_Vars_End          equ *

; Front end code starts here

Start                   di                              ;

Boot                    ld sp,StackTop                  ; Set the stack

                        call InitKeyEntry               ; Setup the "keys"

                        ; Setup the menu entry

                        ld hl,MainMenuDefn              ; Point at it
                        ld (pMainMenu),hl               ;

                        ld a,GF_NormalMasked            ; Default game flags
                        ld (UserGameFlags),a            ;

                        ; Setup the options %%%

DispMain                di                              ; Start the front end

                        ld a,IntPage/256                ;
                        ld i,a                          ;
                        im 2                            ;
                        ei                              ; EI

                        ld hl,sMainMenu                 ; Show the menu
                        call PrintStr                   ;

                        ; At this point we want to flash a menu entry, etc

                        ld ix,(pMainMenu)               ; Point at the menu entry
                        call InitMenuFlashing           ; Start this one

                        ld iy,#5                        ; Don't check select straight away

MainMenuLp              ld ix,(pMenuDefn)               ; Point at the menu entry
                        ld (pMainMenu),ix               ; Store the definition

                        call FlashLine                  ; Flash the line
                        call CheckForKeyPress           ; Any keys pressed?

                        cp pkUp                         ; Move Up the menus
                        call z MM_Up                    ;
                        cp pkDown                       ; Move Down the menus
                        call z MM_Down                  ;

                        dec iyl                         ; Time to allow select?
                        jr nz MainMenuLp                ; No, skip

                        inc iyl                         ; Retore zero

                        cp pkFire                       ; Select this one
                        jr z MM_Select                  ;
                        cp pkSelect                     ; Select this one
                        jr z MM_Select                  ;

                        jr MainMenuLp                   ; Loop

; Start a menu (IX)

InitMenuFlashing        ld (pMenuDefn),ix               ; Save a pointer to the menu
                        ld hl,(ix+dMenuLine)            ; Point at the line
                        ld (pScreenMenuAttr),hl         ; Save it
                        call GrabLine                   ; Copy the attributes

                        ld hl,1                         ; Start flashing it

                        ret                             ; Done

MM_Up                   ex af,af'                       ; Save
                        call PutLine                    ; Undraw the flash

                        ld ix,(pMenuDefn)               ; Point at the menu
                        ld hl,(ix+dMenuUp)              ; Move Up
MM_Change               push hl                         ;
                        pop ix                          ;
                        call InitMenuFlashing           ; Start flashing it

                        ld iy,5                         ; Menu delay

                        ex af,af'                       ; Restore
                        ret                             ; Done

MM_Down                 ex af,af'                       ; Save
                        call PutLine                    ; Undraw the flash

                        ld ix,(pMenuDefn)               ; Point at the menu
                        ld hl,(ix+dMenuDown)            ; Move Down
                        jr MM_Change                    ; As above

MM_Select               ld hl,(ix+dMenuExec)            ; The routine address
                        jp (hl)                         ;

; Show the highscore table

DispHS                  ld hl,0                         ;
                        ld (GameScore),hl               ;

DispHS1                 xor a                           ;
                        ld (EntryF),a                   ;

                        ld hl,sHighA                    ;
                        call PrintStr                   ;

                        ld ix,HighScores                ;
                        ld a,"1"                        ;
DHSlp                   call PrintHS                    ;

                        inc a                           ;
                        cp "9"+1                        ;
                        jr nz,DHS1                      ;

                        add a,"A"-("9"+1)               ;

DHS1                    cp "H"                          ;
                        jr nz,DHSlp                     ;

                        ld a,(EntryF)                   ;
                        or a                            ;
                        jp nz,AcceptEnt                 ;

                        ld hl,sHighB                    ;
                        call PrintStr                   ;

                        ; See if we want to display main, or play

HS_Wait                 call CheckForKeyPress           ; Get the next keypress

                        cp pkSelect                     ;
                        jp z DispMain                   ;

                        cp pk1                          ;
                        jp z StartAGame                 ;

                        cp pkFire                       ;
                        jr nz HS_Wait                   ; Loop

                        ; Now, play whatever game type we last played

StartAGame              ld a,(bBlitzGame)               ;
                        or a                            ;
                        jp nz RestartBlitzGame          ;

                        jp RestartGame                  ;

; Setup for reading keys

InitKeyEntry            ld a,'A'                        ; Start with A
                        ld (RK_CurrChar),a              ;

                        ret                             ; Done

; Get a debounced key - physical key, that is.

CheckForKeyPress        push ix                         ;
                        push hl                         ;
                        push de                         ;
                        push bc                         ;

                        ; We want the randomize the mazes

                        xor a                           ;
                        ld (IrqFlag),a                  ;

SOK_Frame               call Rand                       ;

                        ld a,(IrqFlag)                  ;
                        or a                            ;
                        jr z SOK_Frame                  ;

                        ; Now, read and debounce the keys L/R/U/D/F

                        ld ix,PhysicalKeyTable          ; Point at them
                        ld e,0                          ; Nothing yet

ScanKeyLp               ld bc,(ix+dKeyAddr)             ; The key flag address

                        ld a,c                          ; Addr = 0?
                        or b                            ;
                        jr z SK_X                       ; Done

                        ld a,(bc)                       ; Read the input
                        or a                            ; Pressed?
                        call nz SK_KeyDown              ; Pressed
                        call z SK_KeyUp                 ; Not pressed

                        ; Is this key pressed?

                        ld a,(ix+dKeyPressed)           ; Pressed?
                        or a                            ;
                        jr z SK_MightReport             ; No, skip

                        ; Drive the key repeat

                        ld a,(ix+dKeyRepeat)            ; Dec this
                        dec a                           ;
                        jr nz SK_SaveRepeat             ; Don't repeat yet

                        ld (ix+dKeyReport),1            ; Set the report flag

                        ld a,KEY_REPEAT_CNT             ; Set the timer
SK_SaveRepeat           ld (ix+dKeyRepeat),a            ;

SK_MightReport          ld a,e                          ; Any reported yet?
                        or a                            ;
                        jr nz SK_Next                   ; No, skip

                        ld a,(ix+dKeyReport)            ; Report this one?
                        or a                            ;
                        jr z SK_Next                    ; No, skip

                        ld e,(ix+dKeyName)              ; Report it
                        ld (ix+dKeyReport),0            ; Clear the flag

SK_Next                 ld bc,lKeyEntry                 ; Next
                        add ix,bc                       ;

                        jr ScanKeyLp                    ; Loop

SK_X                    ld a,e                          ; Key the keystroke, if any
                        or a                            ; NZ

                        pop bc                          ;
                        pop de                          ;
                        pop hl                          ;
                        pop ix                          ;
                        ret                             ;

                ; Pressed

SK_KeyDown              push af                         ; Save

                        ld a,(ix+dKeyPressed)           ; Is it pressed?
                        or a                            ;
                        jr nz SK_KD_X                   ; Yes, skip

                        ; Not pressed, inc the counter

                        ld a,(ix+dKeyCount)             ; Inc the counter
                        inc a                           ;
                        ld (ix+dKeyCount),a             ;

                        cp KEY_DEB                      ; Debounced?
                        jr c SK_KD_X                    ; Not yet, exit

                        ; Now debounced

                        ld (ix+dKeyPressed),1           ; Set the pressed flag
                        ld (ix+dKeyReport),1            ; And the report flag
                        ld (ix+dKeyRepeat),iKEY_REPEAT_CNT ;

SK_KD_X                 pop af                          ; Restore
                        ret                             ;

                ; Not pressed

SK_KeyUp                push af                         ; Save

                        ld a,(ix+dKeyPressed)           ; Is it pressed?
                        or a                            ;
                        jr z SK_KU_X                    ; No, skip

                        ; Pressed, dec the counter

                        ld a,(ix+dKeyCount)             ; Dec the counter
                        dec a                           ;
                        ld (ix+dKeyCount),a             ;
                        jr nz SK_KU_X                   ; Not yet, exit

                        ; Now debounced

                        ld (ix+dKeyPressed),0           ; Clear the pressed flag

SK_KU_X                 pop af                          ; Restore
                        ret                             ;

; Get a keystroke, this shows the character and uses up/down/left/right + fire

ReadKeyText             ld a,1                          ;
                        ld (CursorSpeed),a              ;
                        ld (CursorType),a               ;

ReadKeyLp               call CursorInv                  ;

                        call CheckForKeyPress           ;
                        or a                            ;
                        jr z,ReadKeyLp                  ; Nothing, loop

                        push af                         ; Save
                        call CursorOff                  ; Undraw it
                        pop af                          ; Restore

                        ; What have we got?

                        cp pkUp                         ; Inc it
                        jr z RK_IncChar                 ;
                        cp pkDown                       ; Dec it
                        jr z RK_DecChar                 ;
                        cp pkLeft                       ; Delete it
                        jr z RK_Delete                  ;
                        cp pkRight                      ; Enter it
                        jr z RK_Use                     ;
                        cp pkFire                       ; Enter it
                        jr z RK_Enter                   ;

                        jr ReadKeyLp                    ; Ignore

                ; Inc the char

RK_IncChar              ld a,(RK_CurrChar)              ; Next char
                        inc a                           ;
                        cp #80                          ;
                        jr nz RK_SetChar                ;

                        ld a,#20                        ; Wrap
RK_SetChar              ld (RK_CurrChar),a              ;
                        jr ReadKeyText                  ; Done

                ; Dec the char

RK_DecChar              ld a,(RK_CurrChar)              ; Next char
                        dec a                           ;
                        cp #1F                          ;
                        jr nz RK_SetChar                ;

                        ld a,#7F                        ; Wrap
                        jr RK_SetChar                   ;

                ; Delete

RK_Delete               ld a,cDel                       ; Delete
                        ret                             ; Done

                ; Enter

RK_Enter                ld a,cCret                      ; Enter
                        ret                             ; Done

RK_Use                  ld a,(RK_CurrChar)              ; Return this
                        ret                             ; Done

; Fall into...

CursorOff               ld a,(CursorType)               ;
                        or a                            ;
                        jr z,FlipCur                    ;
                        ret                             ;

CursorInv               ld a,(CursorSpeed)              ;
                        dec a                           ;
                        ld (CursorSpeed),a              ;
                        ret nz                          ;

                        ld a,FlashRate                  ;
                        ld (CursorSpeed),a              ;

                        ld a,(CursorType)               ;
                        xor $01                         ;
                        ld (CursorType),a               ;

                        ; XOR the current character onto the screen as a cursor

FlipCur                 ld a,(RK_CurrChar)              ; Get the current char
                        ld l,a                          ;

                        jp pOutChar                     ;

ScanKeys                equ *

SK_Frame                ld a,(IrqFlag)                  ;
                        or a                            ;
                        jr z SK_Frame                   ;

                        xor a                           ;
                        ld (IrqFlag),a                  ;

                        or a                            ;
                        ret                             ;

; Print a character and expand the special ones

PrintSmart              cp cBlurb                       ; Is it a blurb?
                        jp z PS_Blurb                   ; Yes, skip

                        cp cAnEss                       ; Is it an 'S' (last printed number was not 1)
                        jp z PS_AnEss                   ; Yes, skip

                        cp cnThings                     ; Thing-count
                        jr z PS_nThings                 ; Yes, skip

                        cp cnRings                      ; Ring-count
                        jp z PS_nRings                  ; Yes, skip

                        cp cScore                       ; Score
                        jp z PS_Score                   ; Yes, skip

                        cp cTabMid                      ; Align
                        jr z PS_TabMid                  ;

                        cp cTabQuarter                  ; Align
                        jr z PS_TabQuart                ;

                        cp cEnableJokes                 ; Enable them?
                        jp z PS_EnableJokes             ; Yes

                        cp cEnablePiss                  ; Enable them?
                        jp z PS_EnablePiss              ; Yes

                        cp cCAPS                        ; CAPS?
                        jr z PS_CAPS                    ; Yes, skip
                        cp cSYM                         ; SYM?
                        jr z PS_SYM                     ; Yes, skip
                        cp cSPACE                       ; SPACE?
                        jr z PS_SPACE                   ; Yes, skip
                        cp cENTER                       ; ENTER?
                        jr z PS_ENTER                   ; Yes, skip

                        jp Print                        ;

PS_SYM                  call PrintStrFollowing          ;
                        noflow                          ; Tell the assembler this is valid
                        db "SYM",0                      ;
                        ret                             ;

PS_CAPS                 call PrintStrFollowing          ;
                        noflow                          ; Tell the assembler this is valid
                        db "CAPS",0                     ;
                        ret                             ;

PS_ENTER                call PrintStrFollowing          ;
                        noflow                          ; Tell the assembler this is valid
                        db "ENTER",0                    ;
                        ret                             ;

PS_SPACE                call PrintStrFollowing          ;
                        noflow                          ; Tell the assembler this is valid
                        db "SPACE",0                    ;
                        ret                             ;

PS_TabMid               ld a,128                        ;
                        ld (pCurX),a                    ;
                        ret                             ;

PS_TabQuart             ld a,40                         ;
                        ld (pCurX),a                    ;
                        ret                             ;

PS_nThings              push hl                         ;

                        ld hl,(nDTs)                    ;
                        ld a,l                          ;
                        or h                            ;
                        jr z PS_None                    ;

                        ld a,h                          ;
                        xor 1                           ;
                        or l                            ; a=0 if hl=256
                        jr nz PS_Decimal                ;

                        ; We want to say "all"

                        call PrintStrFollowing          ;
                        noflow                          ; Tell the assembler this is valid
                        db "all",0                      ;

                        jr PS_StrPlusal                 ; Set the S flag

                ; None

PS_None                 ld a,1                          ;
                        ld (AnEssFlag),a                ; Set the "S" flag

                        call PrintStrFollowing          ;
                        noflow                          ; Tell the assembler this is valid
                        db "none",0                     ;

PS_StrPlusal            ld a,1                          ;
                        ld (AnEssFlag),a                ; Set the "S" flag

                        pop hl
                        ret

PS_nRings               ld a,(nRings)                   ;

PS_A                    push hl                         ;
                        ld l,a                          ;
                        ld h,0                          ;
                        jr PS_Decimal                   ;

PS_Score                push hl                         ;

                        ld hl,(LastScore)               ; Get the score

                        ; Show the count and set the S flag

PS_Decimal              ld a,l                          ;
                        xor 1                           ; a=0 if l=1
                        or h                            ; a=0 if hl=1
                        ld (AnEssFlag),a                ;

                        call PrintDec_NLZ               ; Print it

                        pop hl                          ; Restore
                        ret                             ; Done

                ; Show an S iff required

PS_AnEss                ld a,(AnEssFlag)                ;
                        or a                            ;
                        ret z                           ;

                        ld a,"s"                        ;
                        jr Print                        ;

                ; Generate a new blurb number

ChooseNewBlurb          ld a,nBlurbs                    ; Get a 0..1-N
                        call RandRange                  ;
                        ld (nBlurb),a                   ;
                        ret                             ; Done

                ; Show a blurb instead

PS_Blurb                push hl                         ; Save
                        push bc                         ;

                        ; Are they valid?

                        ld a,(BlurbValidMarker)         ; Check a very simple flag
                        cp $1F                          ;
                        jr nz PHS_NoBlurbs              ; Overwritten, ignore

                        ; Get the requisite blurb text

                        ld a,(nBlurb)                   ; Get it
                        inc a                           ;
                        ld b,a                          ;

                        ; Find the blurb

                        ld hl,sBlurbs                   ;

PHS_SkipLp              djnz PHS_SkipLp2                ;

                        jr PHS_PrintEncStr              ;

sNoBlurb                dbx *,"Hmmm...",0               ;

                ; There aren't any, show a dummy

PHS_NoBlurbs            ld hl,sNoBlurb                  ;
                        jr PHS_PrintEncStr              ;

                ; Print an encyrpted string

PrintEncStr             push hl                         ;
PHS_PrintEncStr_BC      push bc                         ;

PHS_PrintEncStr         ld c,l                          ;

                        ; Print this entry

PHS_Lp                  ld a,(hl)                       ;
                        call Decrypt_C                  ;
                        jr z PSB_X                      ; Done

                        inc hl                          ;
                        call Print                      ;
                        jr PHS_Lp                       ;

PSB_X                   pop bc                          ;
                        pop hl                          ;
                        ret                             ;

Decrypt_C               xor c                           ;

                        push af                         ;
                        ld a,c                          ;
                        add a,encoff                    ;
                        ld c,a                          ;
                        pop af                          ;
                        ret                             ;

                ; Skip an entry it

PHS_SkipLp2             ld c,l                          ; Decrypt

PHS_SkipLp2Lp           ld a,(hl)                       ;
                        inc hl                          ;
                        call Decrypt_C                  ;
                        jr z PHS_SkipLp                 ; Entry skipped

                        jr PHS_SkipLp2Lp                ;

; Print a character

Print                   push ix                         ;
                        push hl                         ;
                        push de                         ;
                        push bc                         ;
                        push af                         ;

                        cp cWhite+1                     ;
                        jr c,pColour                    ;

                        cp cCls                         ;
                        jr z,pCls                       ;

                        cp cCret                        ;
                        jr z,pCR                        ;

                        cp Home                         ;
                        jr z,pHome                      ;

                        ld h,wChars/256                 ;
                        ld l,a                          ;

                        ld a,(pCurX)                    ;
                        add a,(hl)                      ;
                        call c,pCret                    ;

                        call pOutChar                   ;

                        ld a,(pCurX)                    ;
                        add a,(hl)                      ;
                        ld (pCurX),a                    ;

pExit                   pop af                          ;
                        pop bc                          ;
                        pop de                          ;
                        pop hl                          ;
                        pop ix                          ;
                        ret                             ;

;

pColour                 cp 4                            ;
                        jr nc,pCol1                     ;

                        or $40                          ;

pCol1                   ld (CurCol),a                   ;
                        jr pExit                        ;

pCR                     call pCret                      ;
                        jr pExit                        ;

;

pCls                    ld e,$00                        ;
                        ld hl,Attrs                     ;
                        ld bc,$0300                     ;

pC1                     ld a,(hl)                       ;
                        and $07                         ;
                        jr z,pC2                        ;

                        set 0,e                         ;
                        dec (hl)                        ;
pC2                     cpi                             ;
                        jp pe,pC1                       ;

                        call FrontDelay                 ;

                        bit 0,e                         ;
                        jr nz,pCls                      ;

                        ld hl,ScreenBase                ;
                        ld de,ScreenBase+1              ;
                        ld bc,$1800                     ;

                        xor a                           ;
                        ld (hl),a                       ;

                        ldir                            ;

                        ld a,(CurCol)                   ;
                        ld (hl),a                       ;

                        ld bc,$02FF                     ;

                        ldir                            ;

                        xor a                           ;
                        out ($FE),a                     ;

pHome                   xor a                           ;
                        ld (pCurX),a                    ;
                        ld (ppCurY),a                   ;

                        ld a,cWhite                     ;
                        ld (CurCol),a                   ;

                        ld hl,ScreenBase                ;
                        ld (pCurY),hl                   ;

                        jr pExit                        ;

pCret                   ld a,(ppCurY)                   ;
                        inc a                           ;

SetUpY                  ld (ppCurY),a                   ;

                        and $1F                         ;

                        push hl                         ;

                        ld l,a                          ;
                        ld h,$00                        ;

                        add hl,hl                       ;

                        ld de,YTableFE                  ;
                        add hl,de                       ;

                        ld a,(hl)                       ;
                        inc hl                          ;
                        ld h,(hl)                       ;
                        ld l,a                          ;

                        ld (pCurY),hl                   ;

                        xor a                           ;
                        ld (pCurX),a                    ;

                        pop hl                          ;
                        ret                             ;

pOutChar                bit 7,l                         ;
                        ret nz                          ;

                        push hl                         ;

                        ld h,0                          ;

                        add hl,hl                       ;
                        add hl,hl                       ;
                        add hl,hl                       ;

                        ld de,pChars-256                ;
                        add hl,de                       ;

                        push hl                         ;
                        pop ix                          ;

                        ld a,(pCurX)                    ;

                        rrca                            ;
                        rrca                            ;
                        rrca                            ;
                        and $1F                         ;

                        ld de,(pCurY)                   ;

                        or e                            ;
                        ld e,a                          ;

                        push de                         ;

                        ld a,d                          ;
                        rrca                            ;
                        rrca                            ;
                        rrca                            ;
                        and $03                         ;
                        or $58                          ;
                        ld d,a                          ;

                        ld a,(CurCol)                   ;
                        ld (de),a                       ;
                        inc de                          ;
                        ld (de),a                       ;

                        pop de                          ;

                        ld a,(pCurX)                    ;
                        and $07                         ;
                        ld c,a                          ;

                        ld b,$08                        ;

OutCharL                ld h,(ix)                       ;
                        ld l,$00                        ;

                        ld a,c                          ;
                        or a                            ;
                        jr z,OCL1                       ;

OCL2                    srl h                           ;
                        rr l                            ;
                        dec a                           ;
                        jr nz,OCL2                      ;

OCL1                    ld a,(de)                       ;
                        xor h                           ;
                        ld (de),a                       ;

                        inc e                           ;

                        ld a,(de)                       ;
                        xor l                           ;
                        ld (de),a                       ;

                        dec e                           ;

                        inc ix                          ;
                        inc d                           ;

                        djnz OutCharL                   ;

                        pop hl                          ;
                        ret                             ;

YTableFE                defw $4000,$4020,$4040,$4060    ;
                        defw $4080,$40A0,$40C0,$40E0    ;
                        defw $4800,$4820,$4840,$4860    ;
                        defw $4880,$48A0,$48C0,$48E0    ;
                        defw $5000,$5020,$5040,$5060    ;
                        defw $5080,$50A0,$50C0,$50E0    ;
                        defw 0,0,0,0,0,0,0,0            ;

FrontDelay              push bc                         ;
                        ld bc,$0320                     ;
Del1                    cpi                             ;
                        dec hl                          ;
                        jp pe,Del1                      ;
                        pop bc                          ;
                        ret                             ;

PrintStr                push hl                         ;
                        push af                         ;
PStr1                   ld a,(hl)                       ;
                        inc hl                          ;
                        or a                            ;
                        jr z,PStr2                      ;
                        call PrintSmart                 ;
                        jr PStr1                        ;

PStr2                   pop af                          ;
                        pop hl                          ;
                        ret                             ;

PrintStrFollowing       ex (sp),hl                      ;
                        push af                         ;
PStr11                  ld a,(hl)                       ;
                        inc hl                          ;
                        or a                            ;
                        jr z,PStr12                     ;
                        call PrintSmart                 ;
                        jr PStr11                       ;

PStr12                  pop af                          ;
                        ex (sp),hl                      ;
                        ret                             ;

; Get a text string

GetString               call InitKeyEntry               ;

                        ld ix,HSBuffer                  ;

                        ld a,cMagenta                   ;
                        call Print                      ;

AElp                    push ix
                        call ReadKeyText                ;
                        pop ix

                        or a                            ;
                        jr z,AElp                       ;

                        cp cCret                        ;
                        ret z                           ;

                        cp cDel                         ;
                        jr z,AEDel                      ;

                        cp Clr                          ;
                        jr z,AEclr                      ;

                        cp 8                            ;
                        jr c,AE1                        ;

                        cp $18                          ;
                        jr c,AElp                       ;

AE1                     ld e,a                          ;
                        ld d,wChars/256                 ;
                        ld a,(de)                       ;
                        ld c,a                          ;

                        ld a,(pCurX)                    ;
                        add a,c                         ;
                        cp 255-39                       ;
                        jr nc,AElp                      ;

                        push ix                         ;
                        pop hl                          ;

                        ld bc,HSBuffer                  ;
                        or a                            ;
                        sbc hl,bc                       ;

                        ld bc,130                       ;
                        or a                            ;
                        sbc hl,bc                       ;
                        jr nc,AElp                      ;

                        ld a,e                          ;
                        ld (ix),a                       ;
                        inc ix                          ;
                        call Print                      ;
                        jr AElp                         ;

AEDel                   call AEcDel                     ;
                        jp AElp                         ;

AEclr                   call AEcDel                     ;
                        jr nz,AEclr                     ;

                        jp AElp                         ;

; Remove a character/colour code

AEcDel                  ld de,HSBuffer                  ;

                        push ix                         ;
                        pop hl                          ;

                        or a                            ;
                        sbc hl,de                       ;
                        add hl,de                       ;

                        ret z                           ;

                        dec hl                          ;
                        ld a,(hl)                       ;
                        cp 8                            ;
                        jr c,DelCC                      ;

                        push hl                         ;
                        ld b,a                          ;

                        ld l,a                          ;
                        ld h,wChars/256                 ;

                        ld e,(hl)                       ;
                        ld d,0                          ;

                        ld hl,(pCurX)                   ;
                        or a                            ;
                        sbc hl,de                       ;

                        ld (pCurX),hl                   ;

                        ld a,b                          ;
                        call Print                      ;

                        ld (pCurX),hl                   ;

                        pop ix                          ;
                        or a                            ;
                        ret                             ;

DelCC                   ld de,HSBuffer                  ;
                        ld c,cMagenta                   ;

DCC1                    or a                            ;
                        sbc hl,de                       ;
                        add hl,de                       ;
                        jr z,DCC2                       ;

                        ld a,(de)                       ;
                        cp 8                            ;
                        jr nc,DCC3                      ;
                        ld c,a                          ;

DCC3                    inc de                          ;
                        jr DCC1                         ;

DCC2                    ld a,c                          ;
                        call Print                      ;

                        push hl                         ;
                        pop ix                          ;

                        or a                            ;
                        ret                             ;

AcceptEnt               ld hl,sAccept                   ;
                        call PrintStr                   ;

                        ld hl,(EntryY)                  ;
                        ld (pCurY),hl                   ;

                        ld a,9                          ;
                        ld (pCurX),a                    ;

                        call GetString                  ;

                        push ix                         ;
                        pop hl                          ;

                        call MWEntry                    ; Modify it

                        ld (hl),0                       ; Terminate name
                        inc hl                          ;

                        ld de,(GameScore)               ;
                        ld (hl),e                       ;
                        inc hl                          ;
                        ld a,(GameScore+1)              ;
                        ld (hl),d                       ;
                        inc hl                          ;

                        ld de,HSBuffer                  ;
                        or a                            ;
                        sbc hl,de                       ;

                        ld b,h                          ;
                        ld c,l                          ;
                        push bc                         ;

                        ld hl,(pHSEnd)                  ;
                        ld de,(pInsert)                 ;
                        or a                            ;
                        sbc hl,de                       ;
                        push hl                         ;

                        ld hl,(pHSEnd)                  ;
                        add hl,bc                       ;
                        ex de,hl                        ;
                        ld hl,(pHSEnd)                  ;
                        pop bc                          ;
                        inc bc                          ;
                        lddr                            ;
                        ld hl,HSBuffer                  ;
                        ld de,(pInsert)                 ;
                        pop bc                          ;
                        ldir                            ;

                        ; Now, report the score to the emulator

                        ld hl,(GameScore)               ; Copy store
                        ld (ScoreReportAddr),hl         ; Store it

                        ld hl,ScoreReportVector         ;
                        ld (hl),$C9                     ; RET

                        call ScoreReportVector          ; And call the RET

                        ; Done

                        jp DispHS                       ;

; Mess with the hi-score entry text

MWEntry                 ld de,HSBuffer                  ; Is the entry blank?
                        or a                            ;
                        sbc hl,de                       ;
                        jr nz,MWE1                      ;

                        ; If so we change it to "Someone" for the first 8 tries
                        ; then to "Someone else" after 8 blank entries
                        ; then to "Someone else who is shy or lazy" after 16

                        ld a,(nSomeone)                 ;
                        inc a                           ;
                        ld (nSomeone),a                 ;

                        cp 16                           ; > 16?
                        ld bc,SomeoneEnd-Someone        ;
                        jr nc,MWEa                      ;

                        cp Someone1 - Someone           ; < 8
                        ld bc,8                         ;
                        jr c,MWEa                       ;

                        ld bc,Someone2-Someone          ; 9..15

MWEa                    ld hl,Someone                   ;
                        ld de,HSBuffer                  ;
                        ldir                            ;
                        ex de,hl                        ;
                        ret                             ;

Someone                 db "Someone"                    ;
Someone1                db " else"                      ;
Someone2                db " who is shy or lazy"        ;
SomeoneEnd              equ $                           ;

                ; We have some text

MWE1                    xor a                           ; Clear the someone count
                        ld (nSomeone),a                 ;

                        ld a,l                          ; Entry >= 103 chars?
                        add hl,de                       ;
                        cp 103                          ;
                        jr c,MWE2                       ;

                        ; Replace with the long entry text

                        ld hl,TooLong                   ;
                        ld de,HSBuffer                  ;
                        ld bc,lTooLong                  ;
                        ldir                            ;
                        ex de,hl                        ; pEnd
                        ret                             ;

TooLong                 db "What strange behaviour !"   ;
lTooLong                equ $-TooLong                   ;

MWE2                    ld (hl),0                       ; Terminate the entry
                        push hl                         ; Save HL so we can restore it

                        ld hl,vMazeMasks+4              ;
                        call LoadVectorHL               ;
                        push hl
                        pop ix                          ;

                        ld hl,vMazeMasks+6              ;
                        call LoadVectorHL               ;
                        push hl
                        pop iy                          ;

MWElp                   push ix                         ;

                        ld c,ixl                        ; Encrypt value

                        ld hl,HSBuffer                  ;

Mlp1                    ld a,(hl)                       ; Entry has ended?
                        or a                            ;
                        jr z,Mlpf                       ; Yes, skip

                        cp $21                          ;
                        inc hl                          ;
                        jr c,Mlp1                       ;

                        call ConvLC                     ;

                        ld b,a                          ; Save
                        call GetNextComp_A              ; Get the next byte
                        cp b                            ; Same?
                        jr z,Mlp1                       ; Yes, loop

                        ; They're different. We don't consider this a match.

                        jp MlSkipComp                   ; Skip it (NZ)

        ; The entry has ended, has the comparison?

Mlpf                    call GetNextComp_A              ; Get the next comparison byte
                        or a                            ;

MlSkipComp              pop ix                          ;
                        jr z,MWEcg                      ; Replace the text

                        ; Skip to end of comparison text

                        ld c,ixl                        ; Set the encryption
StepIX                  call GetNextComp_A              ;
                        or a                            ;
                        jr nz,StepIX                    ;

                        ; We have skipped the comparison, so skip the replacement

                        ld c,iyl                        ; Setup the encrypt

StepIY                  call GetNextRepl_A              ;
                        or a                            ;
                        jr nz,StepIY                    ;

                        ; Is this the last comparison text?

                        ld c,ixl                        ; Encrypt
                        ld a,(ix)                       ;
                        xor c                           ;

                        inc a                           ; Is this FF?
                        jr nz,MWElp                     ;

                        pop hl                          ; Restore HL (End of text)
                        ret                             ;

ConvLC                  cp "Z"+1                        ;
                        ret nc                          ;
                        cp "A"                          ;
                        ret c                           ;
                        add a,"a"-"A"                   ;
                        ret                             ;

        ; Replace the text in the highscore buffer

MWEcg                   pop af                          ;

                        ld c,iyl                        ; Setup encrypt

                        ld hl,HSBuffer                  ;

CGlp                    call GetNextRepl_A              ; Get the replacement text

                        or a                            ;
                        ret z                           ;

                        ld (hl),a                       ;
                        inc hl                          ;

                        jr CGlp                         ;

; Get comparison text from (IX) into A

GetNextComp_A           ld a,(ix)                       ; Get it
                        inc ix                          ;

                        xor c                           ; Decrypt

                        push af                         ; Change the encrypt
                        ld a,c                          ;
                        add a,encoff                    ;
                        ld c,a                          ;
                        pop af                          ;

                        ret                             ; Done

; Get replacement text from (IY) into A

GetNextRepl_A           ld a,(iy)                       ; Get it
                        inc iy                          ;

                        xor c                           ; Decrypt

                        push af                         ; Change the encrypt
                        ld a,c                          ;
                        add a,encoff                    ;
                        ld c,a                          ;
                        pop af                          ;

                        ret                             ; Done

GrabLine                ld de,LineBuffer                ;
                        ld bc,32                        ;
                        ldir                            ;
                        ret                             ;

FlashLine               dec hl                          ;

                        ld a,h                          ;
                        or l                            ;
                        ret nz                          ;

                        ld a,(CurType)                  ;
                        inc a                           ;
                        ld (CurType),a                  ;
                        and #01                         ;
                        call z ClearLine                ; Clear it
                        or a
                        call nz PutLine                 ; Restore it

                        ld hl,UDKfr                     ;
                        ret                             ;

PutLine                 ld hl,LineBuffer                ;
                        ld de,(pScreenMenuAttr)         ;
                        ld bc,32                        ;
                        ldir                            ;
                        ret                             ;

ClearLine               ld hl,(pScreenMenuAttr)         ;
                        ld e,l                          ;
                        ld d,h                          ;
                        inc de                          ;
                        ld (hl),$41                     ;
                        ld bc,31                        ;
                        ldir                            ;
                        ret                             ;

ClearGraph              call CGa                        ;

                        inc h                           ;
                        ld a,h                          ;
                        and $07                         ;
                        ret z                           ;

                        jr ClearGraph                   ;

CGa                     push hl                         ;

                        ld d,h                          ;
                        ld e,l                          ;
                        inc de                          ;
                        ld bc,$001F                     ;

                        ld (hl),$00                     ;

                        ldir                            ;

                        pop hl                          ;
                        ret                             ;

DispInst                ld hl,sInst                     ;
DispStatic              call PrintStr                   ;

DIlp                    call CheckForKeyPress           ; Scan the keys
                        or a                            ;
                        jr z,DIlp                       ;

                        jp DispMain                     ;

; Edit the keys

EditKeys                ld ix,EditKeysTable             ; Point at it

EK_Lp                   push ix                         ; Dup
                        pop iy                          ;

                        ld hl,sKeys                     ; Show what we are editing
                        call PrintStr                   ;

                        inc ix                          ; Skip the pointer
                        inc ix                          ;

EK_NameLp               ld a,(ix)                       ; The name
                        inc ix                          ;
                        or a                            ; Done?
                        jr z EK_Show                    ;
                        call PrintSmart                 ;
                        jr EK_NameLp                    ;

EK_Show                 ld a,cCret                      ;
                        call Print                      ;

                        call SavePrintPos               ; Save the print position

                        ; Show the current key until we press something

                        ld hl,(iy+dekMap)               ; Show the present setting
                        call ShowKeyMap                 ;

EK_CheckingCnt          ld c,0                          ;

EK_Check                push bc                         ; Save

                        push iy,ix                      ;
                        call GetCurrKeyBuffer           ; Scan it
                        pop ix,iy                       ;

                        halt                            ;

                        call IsBuffClear                ; Nothing pressed?
                        pop bc                          ; Restore
                        jr nz EK_CheckFailed            ; We've decided to edit

                        inc c                           ; Cnt ++

                        ld a,c                          ; Get the count
                        cp 120                          ;
                        jr nz EK_Check                  ; Not timed out yet

                        jp DispMain                     ; Bugger off...

                ; We're going to edit

EK_CheckFailed          ld hl,(iy+dekMap)               ; Undraw the present setting
                        call ShowKeyMap                 ;

                        ; And edit it

                        call ClearKeyBuff2              ; Clear this one

                        ld a,cWhite                     ;
                        call Print                      ;

EK_UpdateZeroCnt        ld c,0                          ;

EK_Update               push bc                         ; Save

                        push iy,ix                      ;
                        call GetCurrKeyBuffer           ; Scan it
                        pop ix,iy                       ;

                        call ExnorKeyBuffToBuff2        ; Toggle bits

                        ld hl,KeyBuffer2                ; Show the empty one
                        call ShowKeyMap                 ;

EKU_1                   push iy,ix                      ;
                        call GetCurrKeyBuffer           ; Scan it
                        pop ix,iy                       ;

                        call IsBuffClear                ; Something pressed?
                        jr nz EKU_1                     ; Yes, loop

                        call Pause                      ; Wait a bit

                        ld hl,KeyBuffer2                ; Undraw it
                        call ShowKeyMap                 ;

                        call Pause                      ;

                        call CompareKeyBuff2ToBuff3     ; Same?
                        pop bc                          ; Restore
                        jr z EK_Same                    ; Yes, skip

                        call CopyKeyBuff2ToBuff3        ; Copy it
                        jr EK_UpdateZeroCnt             ; Loop

                ; The buffer is the same, has it been the same for long enough?

EK_Same                 inc c                           ; Cnt ++

                        ld a,c                          ; Get the count
                        cp 18                           ;
                        jr z EK_MaybeSet                ;

                        jr EK_Update                    ; Carry on

                ; Stable, is it all clear?

EK_MaybeSet             call IsBuff2Clear               ; Nothing pressed?
                        jp z DispMain                   ; No, exit

                        ld hl,KeyBuffer2                ; Show this key
                        call ShowKeyMap                 ;

                        ; Set the key to this

                        ld hl,KeyBuffer2                ; Copy this set
                        ld de,(iy+dekMap)               ; The keymap
                        ld bc,#5                        ; Length
                        ldir                            ;

                        ; Wait until we release all the keys

EK_NextWait             push iy,ix                      ;
                        call GetCurrKeyBuffer           ; Scan it
                        pop ix,iy                       ;

                        call IsBuffClear                ; Anything pressed?
                        jp nz EK_NextWait               ; Yes, wait

                        ; Now, do we have another?

                        ld a,(ix)                       ; Get the pointer
                        or (ix+1)                       ;
                        jp nz EK_Lp                     ; Next

                        jp DispMain                     ; Top menu

ClearKeyBuff2           ld hl,KeyBuffer2                ; Copy this set
                        ld de,KeyBuffer2+1              ; To here
                        ld (hl),$FF                     ;
                        ld bc,#4                        ; Length
                        ldir                            ;
                        ret                             ; Done

CopyKeyBuff2ToBuff3     ld hl,KeyBuffer2                ; Copy this set
                        ld de,KeyBuffer3                ; To here
                        ld bc,#5                        ; Length
                        ldir                            ;
                        ret                             ; Done

CompareKeyBuff2ToBuff3  ld hl,KeyBuffer2                ; Copy this set
                        ld de,KeyBuffer3                ; To here
                        ld bc,#5                        ; Length
CKBLp                   ld a,(de)                       ;
                        inc de                          ;
                        cpi                             ;
                        ret nz                          ;
                        jp v CKBLp                      ;

                        ret                             ; Done

ExnorKeyBuffToBuff2     ld hl,KeyBuffer                 ; Copy this set
                        ld de,KeyBuffer2                ; To here
                        ld b,#5                         ;

EKBTB2Lp                ld a,(hl)                       ;
                        xor $FF                         ;
                        ld c,a                          ;

                        ld a,(de)                       ;
                        xor c                           ;
                        ld (de),a                       ;

                        inc hl                          ;
                        inc de                          ;
                        djnz EKBTB2Lp                   ;

                        ret                             ; Done

IsBuff2Clear            ld hl,KeyBuffer2                ; Empty?
                        jr IBC1                         ;

IsBuffClear             ld hl,KeyBuffer                 ; Empty?
IBC1                    ld bc,5                         ;

                        ld a,$FF                        ;

IBCLp                   cpi                             ;
                        ret nz                          ;
                        jp v IBCLp                      ;

                        ret                             ; Done

Pause                   ld b,5                          ;
PauseLp                 halt                            ;
                        djnz PauseLp                    ;
                        ret                             ;

SavePrintPos            ld hl,(pCurX)                   ; Save 4 bytes
                        ld (SavepCurX),hl               ;
                        ld hl,(pCurX+2)                 ;
                        ld (SavepCurX+2),hl             ;
                        ret                             ;

LoadPrintPos            ld hl,(SavepCurX)               ; Save 4 bytes
                        ld (pCurX),hl                   ;
                        ld hl,(SavepCurX+2)             ;
                        ld (pCurX+2),hl                 ;
                        ret                             ;

; Show the Spectrum keys which are active - HL is the map

ShowKeyMap              push ix                         ; Save
                        push iy                         ;

                        push hl                         ;
                        pop iy                          ;

                        call LoadPrintPos               ; Restore the print position

                        ld a,cCret                      ;
                        call Print                      ;

                        ld ix,KeyDisplayTable           ; Point at the table

                        ld c,4                          ; 4 rows

SKM_Lp1                 ld b,10                         ; 10 across

SKM_Lp2                 ld a,(ix+1)                     ; The offset
                        push iy                         ;
                        pop hl                          ;

                        add a,l                         ; add the offset
                        ld l,a                          ;
                        adc a,h                         ;
                        sub l                           ;
                        ld h,a                          ;

                        ld a,(hl)                       ; Get the mask

                        or (ix+2)                       ; Mask it
                        inc a                           ; z?
                        jr z SKM_NoShow                 ; Yep, ignore it

                        push bc                         ; Save
                        ld a,(ix)                       ; The name
                        call PrintSmart                 ; Show it

                        ld a," "                        ;
                        call Print                      ;
                        pop bc                          ;

SKM_NoShow              inc ix                          ; Skip this entry
                        inc ix                          ;
                        inc ix                          ;

                        djnz SKM_Lp2                    ;

                        ld a,cCret                      ;
                        call Print                      ;

                        dec c                           ;
                        jr nz SKM_Lp1                   ;

                        pop iy                          ; Restore
                        pop ix                          ;
                        ret                             ;

                        jp DispStatic                   ; As above

PrintHS                 push af                         ;
                        ld (HSChar),a                   ;

                        ld hl,sHighC                    ;
                        call PrintStr                   ;

                        push ix                         ;
                        pop hl                          ;

                        ld a,(EntryF)                   ;
                        or a                            ;
                        jr nz,PHSb                      ;

PHSa                    ld a,(hl)                       ;
                        or a                            ;
                        inc hl                          ;
                        jr nz,PHSa                      ;

                        ld a,(hl)                       ;
                        inc hl                          ;
                        ld h,(hl)                       ;
                        ld l,a                          ;

                        ld de,(GameScore)               ;
                        or a                            ;
                        sbc hl,de                       ;
                        jr nc,PHSb                      ;

                        ld a,1                          ;
                        ld (EntryF),a                   ;

                        ld hl,(pCurY)                   ;
                        ld (EntryY),hl                  ;

                        ld (pInsert),ix                 ;

                        ex de,hl                        ;
                        jp PHSx                         ;

PHSb                    ld a,(ix)                       ;
                        or a                            ;
                        inc ix                          ;
                        jp z,PHS2a                      ;

                        call PrintSmart                 ;
                        jr PHSb                         ;

                ; Show the score

PHS2a                   ld l,(ix)                       ;
                        inc ix                          ;
                        ld h,(ix)                       ;
                        inc ix                          ;
                        ld (pHSEnd),ix                  ;

PHSx                    ld a,255 - (5*6)                ;
                        ld (pCurX),a                    ;

                        ld a,cYellow                    ;
                        call Print                      ;

                        call PrintScoreDec              ;

                        ld a,cCret                      ;
                        call Print                      ;

                        pop af                          ;
                        ret                             ;

PHS2                    inc ix                          ;

PHS3                    call PrintSp                    ;
                        djnz PHS3                       ;

                        jr PHS2a                        ;

PrintSp                 push af                         ;
                        ld a," "                        ;
                        call Print                      ;
                        pop af                          ;
                        ret                             ;

PrintSp2                push af                         ;
                        ld a,$80                        ;
                        call Print                      ;
                        pop af                          ;
                        ret                             ;

PrintDec_NLZ            ld c,1
                        ld de,-10000
                        call PDNL1
                        ld de,-1000
                        call PDNL1
                        ld de,-100
                        call PDNL1
                        ld de,-10
                        call PDNL1
                        LD A,L
                        ADD A,"0"
                        jp Print

PDNL1                   XOR A
PDNL2                   INC A
                        ADD HL,DE
                        JR C,PDNL2
                        SBC HL,DE
                        dec a
                        or "0"

                        CP C
                        JP Z,Print

                        cp "0"
                        ret z

                        ld c,"0"
                        jp Print

PrintScoreDec           ld c,1
                        ld de,-10000
                        call PDD1
                        ld de,-1000
                        call PDD1
                        ld de,-100
                        call PDD1
                        ld de,-10
                        call PDD1
                        LD A,L
                        ADD A,"0"
                        jp Print

PDD1                    XOR A
PDD2                    INC A
                        ADD HL,DE
                        JR C,PDD2
                        SBC HL,DE
                        CP C
                        JR Z,PDDS
                        DEC C
                        ADD A,#2F
                        jp Print

PDDS                    ld a,(pCurX)                    ;
                        add a,6                         ;
                        ld (pCurX),a                    ;
                        ret                             ;

StartGame               ld a,(ix+dMenuFlags)            ; Get the flags
                        ld (UserGameFlags),a            ; Set them

                        ; Fall into

RestartGame             ld a,cCls                       ;
                        call Print                      ;

                        ; Call the game and get the score back

                        call PlayGame                   ; Play a game

                        ; Now, if A=1 we won...

GameOver                call ShowResults                ; Show a report

                        call InitKeyEntry               ; Setup the keys

                        call ChooseNewBlurb             ; Randomise them

                        JP DispHS1                      ; Show the high-scores

; The Blitz game

StartBlitzGame          ld a,(ix+dMenuFlags)            ; Get the flags
                        ld (UserGameFlags),a            ; Set them

                        ; Fall into

RestartBlitzGame        ld a,cCls                       ;
                        call Print                      ;

                        ; Call the game and get the score back

                        call PlayBlitzGame              ; Play a running game

                        ld hl,(GameScore)               ; Halve the score
                        srl h                           ;
                        rr l                            ;
                        ld (GameScore),hl               ;

                        jr GameOver                     ; As above

; Show the results of a game dead/nRings/score

; If won, show appropriate winning text depending on monsters
; If lost show appropriate insult

ShowResults             ld hl,(GameScore)               ; The score
                        ld (LastScore),hl               ; Remember it

                        or a                            ; Won flag?
                        jr nz SR_Won                    ; Yes, skip

                        ; We lost

                        ld hl,sEndLost                  ;
                        call PrintStr                   ;

                        jr SR_Wait                      ; Wait for enter

                ; We won

SR_Won                  ld hl,sEndWon                   ;
                        call PrintStr                   ;

                        ; Add a comment if the options weren't hard enough

                        ld a,(bBlitzGame)               ; Playing blitz game?
                        or a                            ;
                        jr nz SR_NoComment              ;

                        ; Normal game

                        ld a,(UserGameFlags)            ; Easy settings?
                        or a                            ;
                        call z SRW_TryHarder            ; Easy

                        ld a,(UserGameFlags)            ; Are we on the really hard monsters?
                        and gbmMonsterFlags             ; Look only at the monster flags
                        cp gbmMonsterFlags              ; Are they all set?
                        call z SRW_WellDone             ; Full-on

                        ; Enter

SR_NoComment            ld hl,sPressFIRE                ; "Print FIRE"
                        call PrintStr                   ;

                        ; Fall into

SR_Wait                 call CheckForKeyPress           ; Get the next keypress

                        cp pkFire                       ;
                        ret z                           ;
                        cp pk1                          ;
                        ret z                           ;

                        jr SR_Wait                      ; Loop

SRW_TryHarder           ld hl,sTryHarder                ;
                        jp PrintStr                     ;

SRW_WellDone            ld hl,vMazeMasks+2              ;
                        call LoadVectorHL               ;
                        jp PrintEncStr                  ;

; Given HL as a pVector, find the object

LoadVectorHL            push de                         ;

                        ld e,(hl)                       ;
                        inc hl                          ;
                        ld d,(hl)                       ;
                        dec hl                          ;

                        add hl,de                       ;

                        pop de
                        ret                             ;

; Menu text

sEndLost                defb cCls,cCret,cCret,cYellow
                        defb "Hah! You were defeated...",cCret,cCret
                        db cCyan
                        db "You killed ",cnThings," of the things",cCret
                        db "You collected ",cnRings," ring",cAnEss,cCret
                        db "And you somehow scored ",cScore," point",cAnEss,cCret,cCret

                        ; Fall into...

sPressFIRE              db cMagenta,"(Press FIRE)"
                        defb cWhite,0

sTryHarder              defb "Now try it with harder monsters",cCret
                        defb "and/or masked mazes if you dare...",cCret,cCret
                        defb "Use the option menu to set them",cCret,cCret
                        defb 0

sEndWon                 defb cCls,cCret,cCret,cYellow
                        defb "Oi! Neil? Martin? They've beaten our Things...",cCret
                        defb "I told you they weren't nasty enough yet!",cCret,cCret
                        db cCyan
                        db "You killed ",cnThings," of the poor things",cCret
                        db "You collected all the rings and the key",cCret
                        db "And you scored ",cScore," point",cAnEss,cCret,cCret
                        db "I suppose we should say congratulations...",cCret,cCret
                        db 0

mMenuEntry              macro(colour,text)
                          db colour
                          db text
                          db $0D,$0D
                          mend

sMainMenu               db $07,$0C,$86,$86
                        db "Halls Of The Things"
                        db $0D,$0D
                        db cCyan,$86,$86,$82,"(2015 version)"
                        db $0D,$0D,$0D

                        mMenuEntry(cCyan," Play the standard game")
                        mMenuEntry($04," Play blitz-halls (easier)")
                        mMenuEntry($07," Play with masked mazes")
                        mMenuEntry($02," Play on the hardest settings")
                        db $0D
                        mMenuEntry($06," Instructions")
                        mMenuEntry($07," Change The Keys")
                        mMenuEntry($03," HighScores")

                        db $0D
                        mMenuEntry(cCyan,"Use Up/Down to select, then FIRE")
                        db cMagenta,"(c) 1983 .. 2015 Design Design Software"
                        db $00

; Instructions

sInst                   db $0C,$86,$86,$82,"Instructions",$0D,$0D,cCyan
                        db "Search the tower for the seven rings and"
                        db cCret
                        db "try not to get killed while doing it."
                        db cCret
                        db "Then find the key in the bottom level to"
                        db cCret
                        db "escape.",cCret,cCret

                        db "Drink the magic potions (walk over them)"
                        db cCret
                        db "to gain magic. This will automatically heal"
                        db cCret
                        db "wounds and power your weapons."
                        db cCret,cCret
                        db "You don't have to attack the things but"
                        db cCret
                        db "they'll show you no mercy either way."
                        db cCret,cCret
                        db "Use 'select' to align the screen often."
                        db cCret
                        db "Lightning (FIRE) repeat fires and Kickaha will"
                        db cCret
                        db "defend himself with the sword automatically,"
                        db cCret
                        db "but that won't help you much, so good luck..."
                        db cCret,cCret
                        db "Oh - blitz-halls? Fires automatically, and"
                        db cCret
                        db "your reflections can't hurt you. Have fun..."
                        db cCret,cCret,cMagenta
                        db "The source code is available from"
                        db cCret
                        db "www.desdes.com/products/oldfiles"
                        db 0

; Keys

sKeys                   db $0C,$86,$82,cWhite," Set control Keys",$0D,$0D,cCyan
                        db cCret
                        db "You can choose to have more than one key for"
                        db cCret
                        db "each function. And share them."
                        db cCret
                        db "To exit, release all keys and wait."
                        db cCret,cCret,"Press a key to toggle it in the set of"
                        db cCret,"keys that control:  "
                        db cYellow,0

sHighA                  defb $0C,$86,$86,"Halls Of The Things",$0D,$0D
                        defb $86,$86,$83,$05,"Hall of Fame",$0D,$0D,$0D,$00

sHighB                  defb $0D,$0D,"  Press ",$90,$05,"FIRE ",$06
                        defb "to play or ",$05,"SELECT ",$06,"for options",$00

sAccept                 defb $0D,$0D,$85,"Please type in your entry.",$03,$00

; Main menu

gfbNormalSpeed          equ 0                           ; If not set, game runs slowly
gfbMonFaster            equ 1                           ; If set, the monsters are faster
gfbMonDynamic           equ 2                           ; If set, they speed up
gfbMaskedMazes          equ 3

GF_Easy                 equ %0000
GF_NormalMasked         equ %1001
GF_Hard                 equ %1111

MainMenuDefn            equ MainMenuDefn_Standard
MainMenuDefn_Standard   mDefineMenu( 5,MainMenuDefn_Standard,MainMenuDefn_Blitz,StartGame,GF_Easy)
MainMenuDefn_Blitz      mDefineMenu( 7,MainMenuDefn_Standard,MainMenuDefn_Masked,StartBlitzGame,GF_Hard)
MainMenuDefn_Masked     mDefineMenu( 9,MainMenuDefn_Blitz,MainMenuDefn_Hard,StartGame,GF_NormalMasked)
MainMenuDefn_Hard       mDefineMenu(11,MainMenuDefn_Masked,MainMenuDefn_Isntr,StartGame,GF_Hard)
MainMenuDefn_Isntr      mDefineMenu(14,MainMenuDefn_Hard,MainMenuDefn_Keys,DispInst,0)
MainMenuDefn_Keys       mDefineMenu(16,MainMenuDefn_Isntr,MainMenuDefn_HighScore,EditKeys,0)
MainMenuDefn_HighScore  mDefineMenu(18,MainMenuDefn_Keys,MainMenuDefn_HighScore,DispHS,0)

pMenuDefn               ds 2                            ;

; Store the default keymaps

mKeyMap                 macro(name)
Addr                      = label("pkc"+name+"_Offset")
kbit                      = (1 << label("pkc"+name+"_BIT")) xor $FF
                          if Addr = 0
                            db kbit,$FF,$FF,$FF,$FF
                          elseif Addr = 1
                            db $FF,kbit,$FF,$FF,$FF
                          elseif Addr = 2
                            db $FF,$FF,kbit,$FF,$FF
                          elseif Addr = 3
                            db $FF,$FF,$FF,kbit,$FF
                          else
                            db $FF,$FF,$FF,$FF,kbit
                          endif
                          mend

                        align 256

KeyMapLeft              mKeyMap(keyLEFT)                ;
KeyMapRight             mKeyMap(keyRIGHT)               ;
KeyMapUp                mKeyMap(keyUP)                  ;
KeyMapDown              mKeyMap(keyDOWN)                ;
KeyMapJustify           mKeyMap(keyJUSTIFY)             ;
KeyMapStatus            mKeyMap(keySTATUS)              ;
KeyMapFireball          mKeyMap(keyFIREBALL)            ;
KeyMapLightning         mKeyMap(keyLIGHTNING)           ;
KeyMapArrow             mKeyMap(keyARROW)               ;
KeyMapSword             mKeyMap(keySWORD)               ;
KeyMapDrop              mKeyMap(keyDROP)                ;
KeyMapClose             mKeyMap(keyCLOSE)               ;
KeyMapPortal            mKeyMap(keyPORTAL)              ;
KeyMapJoke              mKeyMap(keyJOKE)                ;
KeyBuffer               ds 5,$FF                        ;
KeyBuffer2              ds 5,$FF                        ;
KeyBuffer3              ds 5,$FF                        ;

; The editing keys table

        struct
dekMap                  ds 2
dekName                 ds 1
        send

mSetEKTable             macro(name,mask)
                          dw mask
                          db name,0
                          mend

EditKeysTable           mSetEKTable("Left",KeyMapLeft)      ;
                        mSetEKTable("Right",KeyMapRight)    ;
                        mSetEKTable("Up",KeyMapUp)          ;
                        mSetEKTable("Down",KeyMapDown)      ;

                        mSetEKTable("Justify/Centre man",KeyMapJustify)      ;
                        mSetEKTable("Status",KeyMapStatus)      ;

                        mSetEKTable("Fireball",KeyMapFireball);
                        mSetEKTable("Lightning",KeyMapLightning);
                        mSetEKTable("Arrow",KeyMapArrow);
                        mSetEKTable("Sword",KeyMapSword);

                        mSetEKTable("Drop",KeyMapDrop);
                        mSetEKTable("Close",KeyMapClose);

                        dw 0                            ; Terminator

mShowKeyEntry           macro(name,offset,bitpos)
                          db name
                          db offset
                          db (1 << bitpos) xor $FF
                          mend

; The table used to display the keys

KeyDisplayTable         mShowKeyEntry("1",pkc1_Offset,pkc1_BIT)
                        mShowKeyEntry("2",pkc2_Offset,pkc2_BIT)
                        mShowKeyEntry("3",pkc3_Offset,pkc3_BIT)
                        mShowKeyEntry("4",pkc4_Offset,pkc4_BIT)
                        mShowKeyEntry("5",pkc5_Offset,pkc5_BIT)
                        mShowKeyEntry("6",pkc6_Offset,pkc6_BIT)
                        mShowKeyEntry("7",pkc7_Offset,pkc7_BIT)
                        mShowKeyEntry("8",pkc8_Offset,pkc8_BIT)
                        mShowKeyEntry("9",pkc9_Offset,pkc9_BIT)
                        mShowKeyEntry("0",pkc0_Offset,pkc0_BIT)

                        mShowKeyEntry("Q",pkcQ_Offset,pkcQ_BIT)
                        mShowKeyEntry("W",pkcW_Offset,pkcW_BIT)
                        mShowKeyEntry("E",pkcE_Offset,pkcE_BIT)
                        mShowKeyEntry("R",pkcR_Offset,pkcR_BIT)
                        mShowKeyEntry("T",pkcT_Offset,pkcT_BIT)
                        mShowKeyEntry("Y",pkcY_Offset,pkcY_BIT)
                        mShowKeyEntry("U",pkcU_Offset,pkcU_BIT)
                        mShowKeyEntry("I",pkcI_Offset,pkcI_BIT)
                        mShowKeyEntry("O",pkcO_Offset,pkcO_BIT)
                        mShowKeyEntry("P",pkcP_Offset,pkcP_BIT)

                        mShowKeyEntry("A",pkcA_Offset,pkcA_BIT)
                        mShowKeyEntry("S",pkcS_Offset,pkcS_BIT)
                        mShowKeyEntry("D",pkcD_Offset,pkcD_BIT)
                        mShowKeyEntry("F",pkcF_Offset,pkcF_BIT)
                        mShowKeyEntry("G",pkcG_Offset,pkcG_BIT)
                        mShowKeyEntry("H",pkcH_Offset,pkcH_BIT)
                        mShowKeyEntry("J",pkcJ_Offset,pkcJ_BIT)
                        mShowKeyEntry("K",pkcK_Offset,pkcK_BIT)
                        mShowKeyEntry("L",pkcL_Offset,pkcL_BIT)
                        mShowKeyEntry(cENTER,pkcENTER_Offset,pkcENTER_BIT)

                        mShowKeyEntry(cCAPS,pkcCAPS_Offset,pkcCAPS_BIT)
                        mShowKeyEntry("Z",pkcZ_Offset,pkcZ_BIT)
                        mShowKeyEntry("X",pkcX_Offset,pkcX_BIT)
                        mShowKeyEntry("C",pkcC_Offset,pkcC_BIT)
                        mShowKeyEntry("V",pkcV_Offset,pkcV_BIT)
                        mShowKeyEntry("B",pkcB_Offset,pkcB_BIT)
                        mShowKeyEntry("N",pkcN_Offset,pkcN_BIT)
                        mShowKeyEntry("M",pkcM_Offset,pkcM_BIT)
                        mShowKeyEntry(cSYM,pkcSYM_Offset,pkcSYM_BIT)
                        mShowKeyEntry(cSPACE,pkcSPACE_Offset,pkcSPACE_BIT)

;        if high * xor high KeyMapLeft <> 0
;          zeuserror "Page alignment issues with the data above!"
;        endif

GetCurrKeyBuffer        call ScanUDKeys                 ; Get them

                        ld iy,KeyBuffer                 ;
                        ld (iy),e                       ;
                        ld (iy+1),d                     ;
                        ld (iy+2),l                     ;
                        ld (iy+3),h                     ;
                        ld a,ixl                        ;
                        ld (iy+4),a                     ;

                        ret                             ; Done

; Scan the keys

ScanUDKeys              ld bc,$FEFE                     ; First row
                        in e,(c)                        ; Read it

                        ld b,$FD                        ; Next row
                        in d,(c)                        ; Read it

                        ld b,$FB                        ; Next row
                        in l,(c)                        ; Read it

                        ld b,$F7                        ; Next row
                        in h,(c)                        ; Read it

                        ld b,$EF                        ; Next row
                        in a,(c)                        ; Read it
                        ld ixl,a                        ;

                        ld b,$DF                        ; Next row
                        in a,(c)                        ; Read it
                        ld iyl,a                        ;

                        ld b,$BF                        ; Next row
                        in a,(c)                        ; Read it
                        ld iyh,a                        ;

                        ld b,$7F                        ; Next row
                        in a,(c)                        ; Read it
                        ld ixh,a                        ;

                        ; edlh ixl ixh iyl iyh = keyboard inputs

                        ; Now compress these down to 5 bytes

                        ld a,e                          ; Row 1

                        ld c,iyl                        ;

                        rr c                            ;
                        adc a,a                         ;
                        rr c                            ;
                        adc a,a                         ;
                        rr c                            ;
                        adc a,a                         ;

                        ld e,a                          ; Save

                        ld a,d                          ; Row 2

                        rr c                            ;
                        adc a,a                         ;
                        rr c                            ;
                        adc a,a                         ;

                        ld c,iyh                        ;
                        rr c                            ;
                        adc a,a                         ;

                        ld d,a                          ; Save

                        ld a,l                          ; Row 3

                        rr c                            ;
                        adc a,a                         ;
                        rr c                            ;
                        adc a,a                         ;
                        rr c                            ;
                        adc a,a                         ;

                        ld l,a                          ;

                        ld a,h                          ;

                        rr c                            ;
                        adc a,a                         ;

                        ld c,ixh                        ;
                        rr c                            ;
                        adc a,a                         ;
                        rr c                            ;
                        adc a,a                         ;

                        ld h,a                          ;

                        ld a,ixl                        ;

                        rr c                            ;
                        adc a,a                         ;
                        rr c                            ;
                        adc a,a                         ;
                        rr c                            ;
                        adc a,a                         ;

                        ld ixl,a                        ;

                        ; Keys now mapped into e,d,l,h,ixl

                        ld bc,KeyMapLeft                ; Scan the first key
                        call ScanMap                    ;
                        ld bc,KeyLeft                   ;
                        call DWKey                      ;

                        ld bc,KeyMapRight               ;
                        call ScanMap                    ;
                        ld bc,KeyRight                  ;
                        call DWKey                      ;

                        ld bc,KeyMapUp                  ;
                        call ScanMap                    ;
                        ld bc,KeyUp                     ;
                        call DWKey                      ;

                        ld bc,KeyMapDown                ;
                        call ScanMap                    ;
                        ld bc,KeyDown                   ;
                        call DWKey                      ;

                        ld bc,KeyMapJustify             ;
                        call ScanMap                    ;
                        ld bc,KeyJustify                ;
                        call DWKey                      ;

                        ld bc,KeyMapStatus              ;
                        call ScanMap                    ;
                        ld bc,KeyStatus                 ;
                        call DWKey                      ;

                        ld bc,KeyMapFireball            ;
                        call ScanMap                    ;
                        ld bc,KeyFire                   ;
                        call DWKey                      ;

                        ld bc,KeyMapLightning           ;
                        call ScanMap                    ;
                        ld bc,KeyLight                  ;
                        call DWKey                      ;

                        ld bc,KeyMapArrow               ;
                        call ScanMap                    ;
                        ld bc,KeyArrow                  ;
                        call DWKey                      ;

        if bSWORDKEY_SUPPORTED
                        ld bc,KeyMapSword               ;
                        call ScanMap                    ;
                        ld bc,KeySword                  ;
                        call DWKey                      ;
        endif

                        ld bc,KeyMapDrop                ;
                        call ScanMap                    ;
                        ld bc,KeyDrop                   ;
                        call DWKey                      ;

                        ld bc,KeyMapClose               ;
                        call ScanMap                    ;
                        ld bc,KeyClose                  ;
                        call DWKey                      ;

        if bPORTAL_SUPPORTED
                        ld bc,KeyMapPortal              ;
                        call ScanMap                    ;
                        ld bc,KeyPortal                 ;
                        call DWKey                      ;
        endif

        if bJOKEKEY_SUPPORTED
                        ld bc,KeyMapJoke                ;
                        call ScanMap                    ;
                        ld bc,KeyJoke                   ;
                        call DWKey                      ;
        endif

                        ret                             ;

DWKey                   jr z KillDWKey                  ; Skip if not pressed

                        ; It's pressed

                        ld a,(bc)                       ; Has it been reported?
                        or a                            ;
                        ret nz                          ; Yes, exit

                        ; No, set the pressed flag

                        inc a                           ; A = 1
                        ld (bc),a                       ; Set pressed now
                        inc bc                          ;
                        ld (bc),a                       ; Set "has been" pressed

                        ret                             ;

                ; Key not pressed, clear the pressed flag (but not the has been pressed flag)

KillDWKey               xor a                           ;
                        ld (bc),a                       ;
                        ret                             ;

; Returns NZ if key pressed

ScanMap                 ld a,(bc)                       ; Get the first row
                        or e                            ;
                        inc a                           ;
                        ret nz                          ; Pressed
                        inc c                           ;
                        ld a,(bc)                       ; Get the next row
                        or d                            ;
                        inc a                           ;
                        ret nz                          ; Pressed
                        inc c                           ;
                        ld a,(bc)                       ; Get the next row
                        or l                            ;
                        inc a                           ;
                        ret nz                          ; Pressed
                        inc c                           ;
                        ld a,(bc)                       ; Get the next row
                        or h                            ;
                        inc a                           ;
                        ret nz                          ; Pressed
                        inc c                           ;
                        ld a,(bc)                       ; Get the next row
                        or ixl                          ;
                        inc a                           ;
                        ret                             ; Pressed or not pressed

; Maze type probabilities

ConTab                  defb   0,255,200,150,127,100, 60, 20
                        defb   0,150,100, 40, 23, 13,  7,  4
                        defb   0,200,100, 50, 30, 20, 10,  5

; Text buffers

        struct
                        org BigMaze
HSBuffer                ds 128
LineBuffer              ds 32                           ;

LastScore               ds 2                            ;
AnEssFlag               ds 0                            ; A flag
        send

; Include the game code here

                        include "g.inc"                 ;

; We want to tell the emulator about the writeable sections

                        mem_void $0000,$4000                    ; ROM

                        mem_var $4000,$1B00                     ; Screen

                        mem_var Front_Vars_Start,Front_Vars_End-Front_Vars_Start  ; Front-end variables

; Messy stuff at the end

        if ZEUSVER < 36
            zeuserror "Please upgrade Zeus - the latest version can be found at www.desdes.com/products/oldfiles"
          if ZEUSVER < 37
            zeuserror "Sorry, but this version of Zeus can't generate TZX files"
          endif
          if ZEUSVER < 25
            zeuserror "There are problems with the Spectrum emulation in this version of Zeus"
          endif
          if ZEUSVER < 11
            zeuserror "This version of Zeus doesn't understand the DG pseudo-op"
          endif
        endif

profile $4000,$C000
mem_smc $4000,$C000

; Let's generate a tzx file for it, if we can...

        if ZEUSVER >= 36
          if not enabled bCompressForTape
                zeuserror "This game is too large to be loaded from tape uncompressed..."
                // Save the file without a loader
//                output_tzx "NoLoader.tzx","HOTT2015","",$5B00,Save_Limit-$5B00,2,CS_Test
                // Save the file with Design Design's default border colours
//                output_tzx "HOTT2015.tzx","HOTT2015","",$5800,Save_Limit-$5800,1,CS_Test
                // Save it again with horrible ones
//                output_tzx "UglyBorder.tzx","HOTT2015","",$5800,Save_Limit-$5800,1,CS_Test,$00123456
          else
                import_bin "HallsSCR.bin",$4000 // The original screen
//                import_bin "Screen.scr",$4000  // A sparser screen

                // Save the file without a loader
                output_tzx "NoLoader.tzx","HOTT2015","",$5B00,$A300-$5B00,2,$5B00
                // Save the file with Design Design's default border colours
                output_tzx "HOTT2015.tzx","HOTT2015","",$5800,$A300-$5800,1,$5B00
                // Save it again with horrible ones
                output_tzx "UglyBorder.tzx","HOTT2015","",$5800,$A300-$5800,1,$5B00,$00123456

                zeusprint "Make sure that the compressed block ends below $A300, because that's all we're putting in the tzx file!"
          endif
        endif


